import numpy as np
import pandas as pd
from pathlib import Path
import pytest
from magnethub.loss import LossModel, MATERIALS


def test_smoke():
    mdl = LossModel(material="3C92", team="paderborn")
    # dummy B field data (one trajectory with 1024 samples)
    b_wave = np.random.randn(1024) * 200e-3  # mT
    freq = 124062  # Hz
    temp = 58  # °C

    # get loss and estimated H wave
    p, h = mdl(b_wave, freq, temp)
    assert np.isscalar(p), f"p has shape {p.shape}"
    assert h.shape == (1, 1024), f"h has shape {h.shape}"

    # repetition test
    p2, h2 = mdl(b_wave, freq, temp)
    assert np.allclose(p, p2), f"{p} != {p2}"
    assert np.allclose(h, h2), f"{h} != {h2}"


def test_shorter_sequence():
    mdl = LossModel(material="3C92", team="paderborn")
    # dummy B field data (one trajectory with 1024 samples)
    b_wave = np.random.randn(233) * 200e-3  # mT
    freq = 120_000  # Hz
    temp = 77  # °C

    # get scalar power loss
    p, h = mdl(b_wave, freq, temp)

    assert np.isscalar(p), f"p has shape {p.shape}"
    assert h.shape == (1, 1024), f"h has shape {h.shape}"


def test_batch_execution():
    mdl = LossModel(material="3C92", team="paderborn")

    b_waves = np.random.randn(100, 1024) * 200e-3  # mT
    freqs = np.random.randint(100e3, 750e3, size=100)
    temps = np.random.randint(20, 80, size=100)
    p, h = mdl(b_waves, freqs, temps)

    assert p.size == 100, f"{p.size=}"
    assert h.shape == (100, 1024), f"{h.shape=}"


def test_material_availability():
    b_wave = np.random.randn(1024) * 200e-3  # mT
    freq = 124062  # Hz
    temp = 58  # °C

    for m_lbl in MATERIALS:
        mdl = LossModel(material=m_lbl, team="paderborn")
        p, h = mdl(b_wave, freq, temp)
        assert np.isscalar(p), f"p has shape {p.shape}"
        assert h.shape == (1, 1024), f"h has shape {h.shape}"


@pytest.mark.skip(reason="DEBUG test. The test files do not exist except for the author.")
def test_full_accuracy():
    test_ds = pd.read_csv(Path.cwd() / "tests" / "test_files" / "all_data.csv.gzip", dtype={"material": str})
    for m_lbl in MATERIALS:
        mdl = LossModel(material=m_lbl, team="paderborn")
        test_mat_df = test_ds.query("material == @m_lbl")
        p, h = mdl(
            test_mat_df.loc[:, [c for c in test_mat_df if c.startswith("B_t_")]].to_numpy(),
            test_mat_df.loc[:, "freq"].to_numpy(),
            test_mat_df.loc[:, "temp"].to_numpy(),
        )
        rel_err = np.abs(test_mat_df.ploss - p) / test_mat_df.ploss
        print(f"{m_lbl}: avg. rel err {np.mean(rel_err):.5%} 95th quantile {np.quantile(rel_err, 0.95):.5%}")
        # assert avg_rel_err < 0.70, f"Inaccurate for material {m_lbl} with prediction: {np.abs(test_mat_df.ploss - p)/ test_mat_df.ploss} W/m³"


@pytest.mark.skip(reason="DEBUG test. The test files do not exist except for the author.")
def test_full_accuracy_line_by_line():
    test_ds = pd.read_csv(Path.cwd() / "tests" / "test_files" / "all_data.csv.gzip", dtype={"material": str})
    for m_lbl in MATERIALS:
        mdl = LossModel(material=m_lbl, team="paderborn")
        test_mat_df = test_ds.query("material == @m_lbl")
        p_preds_l = []

        for _, row in test_mat_df.iterrows():
            p, h = mdl(
                row.loc[[c for c in test_mat_df if c.startswith("B_t_")]].to_numpy().astype(np.float64),
                row.loc["freq"],
                row.loc["temp"],
            )
            p_preds_l.append(p)
        p = np.array(p_preds_l)
        rel_err = np.abs(test_mat_df.ploss - p) / test_mat_df.ploss
        print(f"{m_lbl}: avg. rel err {np.mean(rel_err):.5%} 95th quantile {np.quantile(rel_err, 0.95):.5%}")


def test_accuracy_slightly():
    test_ds = pd.read_csv(
        Path(__file__).parent / "test_files" / "unit_test_data_ploss_at_450kWpm3.csv", dtype={"material": str}
    )
    for m_lbl in MATERIALS:
        mdl = LossModel(material=m_lbl, team="paderborn")
        test_mat_df = test_ds.query("material == @m_lbl")
        p, h = mdl(
            test_mat_df.loc[:, [c for c in test_mat_df if c.startswith("B_t_")]].to_numpy(),
            test_mat_df.loc[:, "freq"].to_numpy(),
            test_mat_df.loc[:, "temp"].to_numpy(),
        )
        avg_rel_err = np.mean(np.abs(test_mat_df.ploss - p) / test_mat_df.ploss)
        assert (
            avg_rel_err < 0.08
        ), f"Inaccurate for material {m_lbl} with prediction: {np.abs(test_mat_df.ploss - p)/ test_mat_df.ploss} W/m³"



"""The general loss model.

The LossModel class wraps all other teams' models.
It sanitizes user arguments that would be boilerplate code for any team's code.
"""

from pathlib import Path
import magnethub.paderborn as pb
import magnethub.sydney as sy
import numpy as np

L = 1024  # expected sequence length


MATERIALS = [
    "ML95S",
    "T37",
    "3C90",
    "3C92",
    "3C94",
    "3C95",
    "3E6",
    "3F4",
    "77",
    "78",
    "79",
    "N27",
    "N30",
    "N49",
    "N87",
]

MODEL_ROOT = Path(__file__).parent / "models"

TEAMS = {
    "paderborn": pb.MAT2FILENAME,
    "sydney": sy.MAT2FILENAME,
}


class LossModel:
    """LossModel definition."""

    def __init__(self, material="3C92", team="paderborn"):
        self.material = material.upper()
        self.team = team.lower()

        # value checks
        if self.material not in MATERIALS:
            raise ValueError(f"Chosen material '{self.material}' not supported. Must be either {', '.join(MATERIALS)}")
        if self.team not in list(TEAMS.keys()):
            raise ValueError(f"Chosen team '{self.team}' not supported. Must be in {', '.join(TEAMS.keys())}")

        model_file_name = TEAMS[self.team].get(self.material, None)
        if model_file_name is None:
            raise ValueError(f"Team {self.team.capitalize()} does not offer a model for material {self.material}")
        model_path = MODEL_ROOT / self.team / model_file_name

        # load corresponding model
        match self.team:
            case "paderborn":
                self.mdl = pb.PaderbornModel(model_path, self.material)
            case "sydney":
                self.mdl = sy.SydneyModel(model_path, self.material)

    def __call__(self, b_field, frequency, temperature):
        """Evaluate trajectory and estimate power loss.

        Args
        ----
        b_field: (X, Y) array_like
            The magnetic flux density array(s) in T. First dimension X describes the batch size, the second Y
             the time length (will always be interpolated to 1024 samples)
        frequency: scalar or 1D array-like
            The frequency operation point(s) in Hz
        temperature: scalar or 1D array-like
            The temperature operation point(s) in °C
        
        Return
        ------
        p, h: (X,) np.array, (X, Y) np.ndarray
            The estimated power loss (p) in W/m³ and the estimated magnetic field strength (h) in A/m.
        """
        if b_field.ndim == 1:
            b_field = b_field.reshape(1, -1)

        if b_field.shape[-1] != L:
            actual_len = b_field.shape[-1]
            query_points = np.arange(L)
            support_points = np.arange(actual_len) * L / actual_len
            b_field = np.row_stack([np.interp(query_points, support_points, b_field[i]) for i in range(b_field.shape[0])])

        p, h_seq = self.mdl(b_field, frequency, temperature)

        # may interpolate to 1024 samples if h_seq too short
        if h_seq.shape[-1] != L:
            actual_len = h_seq.shape[-1]
            query_points = np.arange(L)
            support_points = np.arange(actual_len) * L / actual_len
            h_seq = np.row_stack([np.interp(query_points, support_points, h_seq[i]) for i in range(h_seq.shape[0])])
        return p, h_seq



"""
File contains the model according to the Paderborn University approach for the magnet challenge.

Source: https://github.com/moetomg/magnet-engine
"""
import torch
import numpy as np
from scipy.signal import savgol_filter

MAT2FILENAME = {
    "3C90": "3C90.pt",
    "3C92": "3C92.pt",
    "3C94": "3C94.pt",
    "3C95": "3C95.pt",
    "3E6": "3E6.pt",
    "3F4": "3F4.pt",
    "77": "77.pt",
    "78": "78.pt",
    "79": "79.pt",
    "T37": "T37.pt",
    "N27": "N27.pt",
    "N30": "N30.pt",
    "N49": "N49.pt",
    "N87": "N87.pt",
    "ML95S": "ML95S.pt",
}

# Material normalization data (1.B 2.H 3.F 4.T 5.dB/dt)
normsDict ={"77":  [[-2.63253458e-19,  7.47821754e-02],
                    [-7.60950004e-18,  1.10664739e+01],
                    [5.24678898e+00,  2.89351404e-01],
                    [5.87473793e+01,  2.40667381e+01],
                    [6.16727829e+00,  3.83645439e+01]],
            "78":  [[5.67033925e-19,  7.22424510e-02],
                    [-1.54283684e-16,  1.15338828e+01],
                    [5.23810768e+00,  2.89979160e-01],
                    [5.87434082e+01,  2.40685291e+01],
                    [6.09561586e+00,  3.81356049e+01]],
            "79":  [[1.70344847e-13,  9.41321492e-02],
                    [-4.54025068e-02,  3.20463941e+01],
                    [5.21954346e+00,  2.66715437e-01],
                    [5.52068787e+01,  2.37196522e+01],
                    [6.77422905e+00,  3.90895233e+01]],
            "N27": [[7.52738469e-19,  7.48951129e-02],
                    [-8.97477366e-17,  1.47606605e+01],
                    [5.24649334e+00,  2.89964765e-01],
                    [5.87355194e+01,  2.40766029e+01],
                    [6.17841434e+00,  3.84738274e+01]],
            "N30": [[1.43320465e-19,  6.56044649e-02],
                    [-1.57874135e-16,  1.09083332e+01],
                    [5.31786680e+00,  2.78960317e-01],
                    [5.86466904e+01,  2.40616817e+01],
                    [7.01255989e+00,  4.09709969e+01]],
            "N49": [[-8.99073580e-19,  8.94479227e-02],
                    [4.15423721e-16,  3.70622618e+01],
                    [5.25545311e+00,  3.00384015e-01],
                    [5.94716339e+01,  2.44349327e+01],
                    [6.75209475e+00,  3.91901703e+01]],
            "N87": [[1.72051200e-13,  6.26231476e-02],
                    [4.02299992e-02,  7.61060358e+00],
                    [5.26309967e+00,  2.87137657e-01],
                    [5.83059006e+01,  2.40639057e+01],
                    [6.53078842e+00,  3.93127785e+01]],
            "3E6": [[1.01579639e-18,  7.04261607e-02],
                    [2.34374135e-16,  7.21573964e+00],
                    [5.34307003e+00,  2.66708523e-01],
                    [5.86578026e+01,  2.40552864e+01],
                    [7.23155785e+00,  4.15975838e+01]],
            "3F4": [[-1.75200068e-19,  5.98892952e-02],
                    [-9.48865199e-18,  4.74414811e+01],
                    [5.14398336e+00,  3.04210454e-01],
                    [5.76523476e+01,  2.43824081e+01],
                    [6.23030663e+00,  3.64991379e+01]],
            "T37": [[1.72051200e-13,  6.26231476e-02],
                    [4.02299992e-02,  7.61060358e+00],
                    [5.26309967e+00,  2.87137657e-01],
                    [5.83059006e+01,  2.40639057e+01],
                    [6.53078842e+00,  3.93127785e+01]],
            "3C90":[[-3.27923689e-19,  6.56109348e-02],
                    [6.99196716e-17,  1.26583787e+01],
                    [5.19875193e+00,  2.68499136e-01],
                    [5.86049919e+01,  2.40574703e+01],
                    [6.29652929e+00,  3.84585190e+01]],
            "3C92":[[-2.35520104e-13,  6.53518693e-02],
                    [1.18689366e-01,  1.23585692e+01],
                    [5.16579533e+00,  2.73998171e-01],
                    [5.84305267e+01,  2.40970516e+01],
                    [5.88209248e+00,  3.69935722e+01]],
            "3C94":[[1.21232679e-19,  7.44383659e-02],
                    [-2.19613879e-17,  1.18042579e+01],
                    [5.22766781e+00,  2.68348873e-01],
                    [5.87128143e+01,  2.40769634e+01],
                    [6.53718996e+00,  3.91955910e+01]],
            "3C95":[[5.64116728e-14,  7.90115297e-02],
                    [1.11898437e-01,  1.29696641e+01],
                    [5.18842697e+00,  2.69014776e-01],
                    [5.86223640e+01,  2.40957470e+01],
                    [6.25767517e+00,  3.84026108e+01]],
           "ML95S":[[-1.53185180e-13,  1.15827541e-01],
                    [3.84426934e-01,  4.45061606e+01],
                    [5.21606445e+00,  2.65364528e-01],
                    [5.70770302e+01,  2.44398289e+01],
                    [7.30377579e+00,  4.04136391e+01]],
             }

# %% Initialize model
class SydneyModel:
    """The Sydney model."""

    def __init__(self, mdl_path, material):
        # Select GPU as default device
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # 1.Create model isntances
        self.mdl = MMINet(material).to(self.device)   
        
        # 2.Load specific model 
        state_dict = torch.load(mdl_path, map_location=self.device)
        self.mdl.load_state_dict(state_dict,strict=True)
        
    
    def __call__(self, data_B, data_F, data_T):
        """Call method."""
        # ----------------------------------------------------------- batch execution  
        # if isinstance(data_F, np.ndarray):

        # 1.Get dataloader
        loader = get_dataloader(data_B,data_F,data_T,self.mdl.norm)

        # 2.Validate the models
        data_P = torch.Tensor([]).to(self.device)  # Allocate memory to store loss density

        with torch.no_grad():
            # Start model evaluation explicitly
            self.mdl.eval()
            for inputs, vars in loader:
                    Pv, h_series = self.mdl(inputs.to(self.device), vars.to(self.device))

                    data_P = torch.cat((data_P,Pv.to(self.device)),dim=0)
        data_P, h_series = data_P.cpu().numpy(), h_series.cpu().numpy()
        if data_P.size == 1:
            data_P = data_P.item()
        if h_series.ndim == 1:
            h_series = h_series.reshape(1, -1)
        return data_P, h_series

class MMINet(torch.nn.Module):
    """
    Magnetization mechanism-determined neural network.

    Parameters:
    - hidden_size: number of eddy current slices (RNN neuron)
    - operator_size: number of operators
    - input_size: number of inputs (1.B 2.dB 3.dB/dt)
    - var_size: number of supplenmentary variables (1.F 2.T)
    - output_size: number of outputs (1.H)
    """

    def __init__(self, Material, hidden_size=30, operator_size=30,
                  input_size=3,var_size=2,output_size=1):
        super().__init__()   
        self.input_size = input_size
        self.var_size = var_size
        self.hidden_size = hidden_size
        self.output_size = output_size 
        self.operator_size = operator_size
        self.norm = normsDict[Material]           # normalization data
        self.n_init = 32

        # Consturct the network 
        self.rnn1 = StopOperatorCell(self.operator_size)
        self.dnn1 = torch.nn.Linear(self.operator_size+2,1)
        self.rnn2 = EddyCell(4,self.hidden_size,output_size)
        self.dnn2 = torch.nn.Linear(self.hidden_size,1)

        self.rnn2_hx = None

    def forward(self, x, var):
        """
        Forward function.

        Parameters:
        x: batch,seq,input_size
            Input features (1.B, 2.dB, 3.dB/dt)
        var: batch,var_size
            Supplementary inputs (1.F 2.T)
        """    
        batch_size = x.size(0)          # Batch size 
        seq_size = x.size(1)            # Ser
        
        # Initialize operator state
        self.rnn1_hx = var[:,2:]

        # Initialize DNN2 input (1.B 2.dB/dt)
        x2 = torch.cat((x[:,:,0:1],x[:,:,2:3]),dim=2)
        for t in range(seq_size):
            # RNN1 input (dB,state)       
            self.rnn1_hx = self.rnn1(x[:,t,1:2], self.rnn1_hx)

            # DNN1 input (rnn1_hx,F,T)
            dnn1_in = torch.cat((self.rnn1_hx,var[:,0:2]),dim=1) 

            # H hysteresis prediction 
            H_hyst_pred = self.dnn1(dnn1_in)

            # DNN2 input (B,dB/dt,T,F)
            rnn2_in = torch.cat((x2[:,t,:],var[:,0:2]),dim=1) 

            # Initialize second rnn state 
            if t==0:
                H_eddy_init = x[:,t,0:1]-H_hyst_pred
                buffer = x.new_ones(x.size(0),self.hidden_size)
                self.rnn2_hx = torch.autograd.Variable((buffer/torch.sum(self.dnn2.weight,dim=1))*H_eddy_init)

            self.rnn2_hx = self.rnn2(rnn2_in, self.rnn2_hx)

            # H eddy prediction
            H_eddy = self.dnn2(self.rnn2_hx)

            # H total 
            H_total = (H_hyst_pred+H_eddy).view(batch_size,1,self.output_size)

            if t==0:
                output = H_total
            else:
                output = torch.cat((output,H_total),dim=1)

        # Compute the power loss density 
        B = (x[:,self.n_init:,0:1]*self.norm[0][1]+self.norm[0][0]) 
        H = (output[:, self.n_init:,:]*self.norm[1][1]+self.norm[1][0])
        Pv = torch.trapz(H, B, axis=1)*(10**(var[:, 0:1]*self.norm[2][1]+self.norm[2][0]))

        # Return results
        H = savgol_filter(H.detach().to("cpu").numpy(), window_length=7, polyorder=2,axis=1)
        H = torch.from_numpy(H).view(batch_size,-1,1)
        real_H = torch.cat((H[:, -self.n_init:, :],H[:, :-self.n_init, :]), dim=1)
        return torch.flatten(Pv).cpu(), real_H[:, :, 0].cpu()

class StopOperatorCell():
    """
    MMINN Sub-layer: Static hysteresis prediction using stop operators.

    Parameters:
    - operator_size: number of operator
    """

    def __init__(self, operator_size):
        self.operator_thre = torch.pow(torch.arange(1, operator_size+1, dtype=torch.float)/(operator_size+1), torch.tensor(3.0)).view(1,-1)*1
        
    def sslu(self, X):
        """Hardsimoid-like or symmetric saturated linear unit definition."""
        a = torch.ones_like(X)
        return torch.max(-a,torch.min(a,X))
    
    def __call__(self, dB, state):
        """Update operator of each time step."""
        r = self.operator_thre.to(dB.device) 
        output = self.sslu((dB + state)/r)*r
        return output.float()


class EddyCell(torch.nn.Module):
    """
    MMINN subsubnetwork: Dynamic hysteresis prediction.

    Parameters:
    - input_size: feature size
    - hidden_size: number of hidden units (eddy current layers)
    - output_size: number of the output
    """

    def __init__(self, input_size, hidden_size, output_size=1):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size

        self.x2h = torch.nn.Linear(input_size, hidden_size ,bias=False)
        self.h2h = torch.nn.Linear(hidden_size, hidden_size ,bias=False)

    def forward(self, x, hidden=None):
        """
        Forward function.

        Parameters:
        x: batch,input_size
            features (1.B 2.dB/dt 3.F 4.T)
        hidden: batch,hidden_size
            dynamic hysteresis effects at each eddy current layer
        """
        hidden = self.x2h(x) + self.h2h(hidden)
        hidden = torch.sigmoid(hidden)
        return hidden


def get_dataloader(data_B, data_F, data_T, norm, n_init=32):
    """
    Preprocess data into a data loader.

    Get a test dataloader.

    Parameters
    ---------
    data_B: array
         B data
    data_F
         F data
    data_T
         T data
    norm : list 
         B/F/T normalization data
    n_init : int
         Additional points for computing the history magnetization
    """
    # Data pre-process 
    # 1. Down-sample to 128 points 
    seq_length = 128
    cols = np.array(range(0,1023,8))
    data_B = data_B[:, cols]

    # 2. Add extra points for initial magnetization calculation 
    data_length = seq_length + n_init
    data_B = np.hstack((data_B,data_B[:,1:1+n_init]))

    # 3. Format data into tensors 
    B = torch.from_numpy(data_B).view(-1, data_length, 1).float()
    if np.isscalar(data_F):
        data_F = np.array([data_F])
    if np.isscalar(data_T):
        data_T = np.array([data_T])
    T = torch.from_numpy(data_T).view(-1, 1).float()
    F = torch.from_numpy(np.log10(data_F)).view(-1,1).float()

    # 4. Data Normalization 
    in_B = (B-norm[0][0])/norm[0][1]
    in_F = (F-norm[2][0])/norm[2][1]
    in_T = (T-norm[3][0])/norm[3][1]

    # 5. Extra features 
    in_dB = torch.diff(in_B,dim=1)                     # Flux density change
    in_dB = torch.cat((in_dB[:, 0:1, :], in_dB), dim=1)
    
    dB_dt = in_dB*(seq_length*F.reshape(-1, 1, 1))
    in_dB_dt = (dB_dt-norm[4][0])/norm[4][1]        # Flux density change rate
    
    max_B, _ = torch.max(in_B, dim=1)
    min_B, _ = torch.min(in_B, dim=1)

    s0 = get_operator_init(in_B[:, 0, 0]-in_dB[:, 0, 0], in_dB, max_B, min_B)  # Operator inital state

    # 6. Create dataloader to speed up data processing
    test_dataset = torch.utils.data.TensorDataset(torch.cat((in_B, in_dB, in_dB_dt), dim=2), torch.cat((in_F, in_T, s0),dim=1))
    kwargs = {'num_workers': 0, 'batch_size': 128, 'drop_last': False}
    test_loader = torch.utils.data.DataLoader(test_dataset, **kwargs)

    return test_loader


# %% Predict the operator state at t0
def get_operator_init(B0, dB, Bmax, Bmin, operator_size=30, max_out_H=1):
    """Compute the initial state of hysteresis operators.

    Parameters
    ---------
    B0 : torch_like (batch)
         Stop operator excitation at t1
    dB : torch_like (batch, data_length)
         Flux density changes at each t
    Bmax: torch_like (batch)
         Max flux density of each cycle
    Bmin: torch_like (batch)
         Min flux density of each cycle
    operator_size: int
         TBD
    max_out_H:
         TBD
    """
    # 1. Parameter setting
    batch = dB.shape[0]
    state = torch.zeros((batch, operator_size))
    operator_thre = torch.pow(torch.arange(1, operator_size+1, dtype=torch.float)/operator_size+1, torch.tensor(3.0)).view(1, -1)*max_out_H

    # 2. Iterate each excitation for the operator inital state computation
    for i in range(B0.__len__()):
        for j in range(operator_size):
            r = operator_thre[0, j]
            if (Bmax[i] >= r) or (Bmin[i] <= -r):
                if dB[i, 0] >= 0:
                    if B0[i] > Bmin[i]+2*r:
                        state[i, j] = r
                    else:
                        state[i, j] = B0[i]-(r+Bmin[i])
                else:
                    if B0[i] < Bmax[i]-2*r:
                        state[i, j] = -r
                    else:
                        state[i, j] = B0[i]+(r-Bmax[i])

    return state



"""
File contains the model according to the Paderborn University approach for the magnet challenge.

Source: https://github.com/upb-lea/hardcore-magnet-challenge
"""

import numpy as np
import pandas as pd
import torch


L = 1024  # expected sequence length
ALL_B_COLS = [f"B_t_{k}" for k in range(L)]
ALL_H_COLS = [f"H_t_{k}" for k in range(L)]
FREQ_SCALE = 150_000.0  # in Hz

# material constants
MAT_CONST_B_MAX = {
    "3C90": 0.282254066096809,
    "3C94": 0.281179823717941,
    "3E6": 0.199842551960829,
    "3F4": 0.312954906128548,
    "77": 0.315644038162322,
    "78": 0.3166405692215,
    "3C92": 0.319193837507623,
    "T37": 0.253934182092085,
    "3C95": 0.322678797082694,
    "79": 0.314715273611617,
    "ML95S": 0.330102949741973,
    "N27": 0.317335585296054,
    "N30": 0.201167700159802,
    "N49": 0.317828937072173,
    "N87": 0.280909134946228,
}  # in T
MAT_CONST_H_MAX = {
    "3C90": 84.7148502254261,
    "3C94": 64.8575649838852,
    "3E6": 74.1579701817075,
    "3F4": 150,
    "77": 86.5681744566843,
    "78": 87.5896894086919,
    "3C92": 150,
    "T37": 87.7490689795367,
    "3C95": 72.0625845199264,
    "79": 150,
    "ML95S": 150,
    "N27": 119.039616554254,
    "N30": 116.951204964406,
    "N49": 150,
    "N87": 100.674197407678,
}  # in A/m

# Normalization constants for misc. input features (absolute maximum over entire data set)
NORM_DENOM = {
    "3C90": {
        "b_peak2peak": 0.562775,
        "log_peak2peak": 3.854492,
        "mean_abs_dbdt": 0.001098,
        "log_mean_abs_dbdt": 10.097290,
        "sample_time": 0.000020,
        "b_deriv": 9888.829113924028,
        "b_deriv_sq": 677552639.6192261,
    },
    "3C92": {
        "b_peak2peak": 0.638772,
        "log_peak2peak": 3.952986,
        "mean_abs_dbdt": 0.001246,
        "log_mean_abs_dbdt": 10.191507,
        "sample_time": 0.000020,
        "b_deriv": 8367.455279302658,
        "b_deriv_sq": 352400581.9367888,
    },
    "3C94": {
        "b_peak2peak": 0.561696288415998,
        "log_peak2peak": 3.8527982020339397,
        "mean_abs_dbdt": 0.001094609016982872,
        "log_mean_abs_dbdt": 10.09547460054903,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 9369.848364186466,
        "b_deriv_sq": 722100525.0000031,
    },
    "3C95": {
        "b_peak2peak": 0.643972929264389,
        "log_peak2peak": 3.951775170875342,
        "mean_abs_dbdt": 0.001255971974873893,
        "log_mean_abs_dbdt": 10.19012406875038,
        "sample_time": 2.000400080016003e-05,
        "b_deriv": 8434.886044920588,
        "b_deriv_sq": 394837993.91482246,
    },
    "3E6": {
        "b_peak2peak": 0.39860703974419,
        "log_peak2peak": 3.9943679877190745,
        "mean_abs_dbdt": 0.0007754383944842351,
        "log_mean_abs_dbdt": 10.237231774124638,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10961.711420802148,
        "b_deriv_sq": 594537528.5676531,
    },
    "3F4": {
        "b_peak2peak": 0.62579,
        "log_peak2peak": 3.966479191224205,
        "mean_abs_dbdt": 0.0012214485516744327,
        "log_mean_abs_dbdt": 10.204883300750092,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 11289.048646402262,
        "b_deriv_sq": 767625384.1351097,
    },
    "77": {
        "b_peak2peak": 0.62941,
        "log_peak2peak": 3.969122597468719,
        "mean_abs_dbdt": 0.001227556207233634,
        "log_mean_abs_dbdt": 10.207264569238735,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10512.189743589744,
        "b_deriv_sq": 705767456.3398341,
    },
    "78": {
        "b_peak2peak": 0.6315575499770669,
        "log_peak2peak": 3.9789526676175275,
        "mean_abs_dbdt": 0.001231771095979432,
        "log_mean_abs_dbdt": 10.217365718778147,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10374.080553295375,
        "b_deriv_sq": 688850200.676364,
    },
    "79": {
        "b_peak2peak": 0.648230254830443,
        "log_peak2peak": 3.9557162612448136,
        "mean_abs_dbdt": 0.0012638156025685287,
        "log_mean_abs_dbdt": 10.194239618633777,
        "sample_time": 2.00160128102482e-05,
        "b_deriv": 8233.983883950554,
        "b_deriv_sq": 495462859.8505281,
    },
    "T37": {
        "b_peak2peak": 0.509931827049285,
        "log_peak2peak": 3.9681646195793543,
        "mean_abs_dbdt": 0.0009933972666621286,
        "log_mean_abs_dbdt": 10.207037964928544,
        "sample_time": 2.00160128102482e-05,
        "b_deriv": 10816.309729628758,
        "b_deriv_sq": 757546320.923036,
    },
    "N27": {
        "b_peak2peak": 0.6318699999999999,
        "log_peak2peak": 3.96936864553962,
        "mean_abs_dbdt": 0.0012323949169110505,
        "log_mean_abs_dbdt": 10.207788696606462,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10348.1536885246,
        "b_deriv_sq": 655163308.8104846,
    },
    "N30": {
        "b_peak2peak": 0.40126,
        "log_peak2peak": 3.9675357153584048,
        "mean_abs_dbdt": 0.0007812218963831858,
        "log_mean_abs_dbdt": 10.20594094230823,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10772.915082382753,
        "b_deriv_sq": 640974360.0000046,
    },
    "N49": {
        "b_peak2peak": 0.6349207696798069,
        "log_peak2peak": 3.729801733358157,
        "mean_abs_dbdt": 0.0012362016813373098,
        "log_mean_abs_dbdt": 9.972437816965455,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10407.538775510218,
        "b_deriv_sq": 742250594.1211493,
    },
    "N87": {
        "b_peak2peak": 0.561429326791338,
        "log_peak2peak": 3.8619698511811014,
        "mean_abs_dbdt": 0.001094995270801449,
        "log_mean_abs_dbdt": 10.104824015905193,
        "sample_time": 2.002002002002002e-05,
        "b_deriv": 10006.669716376984,
        "b_deriv_sq": 616830982.1245888,
    },
    "ML95S": {
        "b_peak2peak": 0.660470219971137,
        "log_peak2peak": 3.684324864126149,
        "mean_abs_dbdt": 0.0012853467115523114,
        "log_mean_abs_dbdt": 9.922533485496238,
        "sample_time": 1.997602876548142e-05,
        "b_deriv": 8000.255232581114,
        "b_deriv_sq": 434328441.4934993,
    },
}

MAT2FILENAME = {
    "3C90": "cnn_3C90_experiment_1b4d8_model_f3915868_seed_0_fold_0.pt",
    "3C92": "cnn_3C92_experiment_ea1fe_model_72510647_seed_0_fold_0.pt",
    "3C94": "cnn_3C94_experiment_56441_model_55693612_seed_0_fold_0.pt",
    "3C95": "cnn_3C95_experiment_ad6ec_model_046383a5_seed_0_fold_0.pt",
    "3E6": "cnn_3E6_experiment_a7817_model_a1035e58_seed_0_fold_0.pt",
    "3F4": "cnn_3F4_experiment_c234d_model_2d43b97d_seed_0_fold_0.pt",
    "77": "cnn_77_experiment_268ae_model_5b7c92ed_seed_0_fold_0.pt",
    "78": "cnn_78_experiment_e5297_model_77fbd758_seed_0_fold_0.pt",
    "79": "cnn_79_experiment_45989_model_7227af72_seed_0_fold_0.pt",
    "T37": "cnn_T37_experiment_a084a_model_fb31325e_seed_0_fold_0.pt",
    "N27": "cnn_N27_experiment_8b7f0_model_20954a2a_seed_0_fold_0.pt",
    "N30": "cnn_N30_experiment_5a78c_model_6bd86623_seed_0_fold_0.pt",
    "N49": "cnn_N49_experiment_27442_model_d6234a32_seed_0_fold_0.pt",
    "N87": "cnn_N87_experiment_985cb_model_d51b0f7e_seed_0_fold_0.pt",
    "ML95S": "cnn_ML95S_experiment_1c978_model_883a5d2f_seed_0_fold_0.pt",
}


def form_factor(x):
    """
    Calculate the form factor.

    definition:      kf = rms(x) / mean(abs(x))
    for ideal sine:  np.pi/(2*np.sqrt(2))
    """
    return np.sqrt(np.mean(x**2, axis=1)) / np.mean(np.abs(x), axis=1)


def crest_factor(x):
    """
    Calculate the crest factor.

    definition:      kc = rms(x) / max(x)
    for ideal sine:  np.sqrt(2)
    """
    return np.max(np.abs(x), axis=1) / np.sqrt(np.mean(x**2, axis=1))


def bool_filter_sine(b, rel_kf=0.01, rel_kc=0.01, rel_0_dev=0.1):
    """
    Bool classification for sinusoidal waveforms, used by function get_waveform_est.

    b: input flux density (nxm)-array with n m-dimensional flux density waveforms
    rel_kf: (allowed) relative deviation of the form factor for sine classification
    rel_kc: (allowed) relative deviation of the crest factor for sine classification
    rel_0_dev: (allowed) relative deviation of the first value from zero (normalized on the peak value)
    """
    kf_sine = np.pi / (2 * np.sqrt(2))
    kc_sine = np.sqrt(2)

    b_ff = form_factor(b)
    b_cf = crest_factor(b)
    b_max = np.max(b, axis=1)
    mask = np.all(
        np.column_stack(
            [
                b_ff < kf_sine * (1 + rel_kf),  # form factor based checking
                b_ff > kf_sine * (1 - rel_kf),  # form factor based checking
                b_cf < kc_sine * (1 + rel_kc),  # crest factor based checking
                b_cf > kc_sine * (1 - rel_kc),  # crest factor based checking
                b[:, 0] < b_max * rel_0_dev,  # starting value based checking
                b[:, 0] > -b_max * rel_0_dev,  # starting value based checking
            ]
        ),
        axis=1,
    )

    return mask


def bool_filter_triangular(b, rel_kf=0.005, rel_kc=0.005):
    """Bool classification for triangular waveforms, used by function get_waveform_est."""
    kf_triangular = 2 / np.sqrt(3)
    kc_triangular = np.sqrt(3)

    b_ff = form_factor(b)
    b_cf = crest_factor(b)

    mask = np.all(
        np.column_stack(
            [
                b_ff < kf_triangular * (1 + rel_kf),
                b_ff > kf_triangular * (1 - rel_kf),
                b_cf < kc_triangular * (1 + rel_kc),
                b_cf > kc_triangular * (1 - rel_kc),
            ]
        ),
        axis=1,
    )

    return mask


def get_waveform_est(full_b):
    """
    Classify the waveforms into [other, square, triangular, sine].

    From Till's tp-1.4.7.3.1 NB, return waveform class.
    Postprocessing from wk-1.1-EDA NB.

    Return class estimate 'k', where [0, 1, 2, 3] corresponds to
    [other, square, triangular, sine].
    """
    # labels init all with 'other'
    k = np.zeros(full_b.shape[0], dtype=int)

    # square
    k[
        np.all(
            np.abs(full_b[:, 250:500:50] - full_b[:, 200:450:50]) / np.max(np.abs(full_b), axis=1, keepdims=True)
            < 0.05,
            axis=1,
        )
        & np.all(full_b[:, -200:] < 0, axis=1)
    ] = 1

    # triangular
    k[bool_filter_triangular(full_b, rel_kf=0.01, rel_kc=0.01)] = 2

    # sine
    k[bool_filter_sine(full_b, rel_kf=0.01, rel_kc=0.01)] = 3

    # postprocess "other" signals in frequency-domain, to recover some more squares, triangles, and sines
    n_subsample = 32
    other_b = full_b[k == 0, ::n_subsample]
    other_b /= np.abs(other_b).max(axis=1, keepdims=True)
    other_b_ft = np.abs(np.fft.fft(other_b, axis=1))
    other_b_ft /= other_b_ft.max(axis=1, keepdims=True)
    msk_of_newly_identified_sines = np.all((other_b_ft[:, 3:10] < 0.03) & (other_b_ft[:, [2]] < 0.2), axis=1)
    msk_of_newly_identified_triangs = np.all(((other_b_ft[:, 1:8] - other_b_ft[:, 2:9]) > 0), axis=1) | np.all(
        ((other_b_ft[:, 1:8:2] > 1e-2) & (other_b_ft[:, 2:9:2] < 1e-2)), axis=1
    )
    msk_of_newly_identified_triangs = msk_of_newly_identified_triangs & ~msk_of_newly_identified_sines
    msk_of_newly_identified_squares = np.all((other_b_ft[:, 1:4:2] > 1e-2) & (other_b_ft[:, 2:5:2] < 1e-3), axis=1)
    msk_of_newly_identified_squares = (
        msk_of_newly_identified_squares & ~msk_of_newly_identified_sines & ~msk_of_newly_identified_triangs
    )
    idx_sines = np.arange(k.size)[k == 0][msk_of_newly_identified_sines]
    idx_triangs = np.arange(k.size)[k == 0][msk_of_newly_identified_triangs]
    idx_squares = np.arange(k.size)[k == 0][msk_of_newly_identified_squares]
    k[idx_squares] = 1
    k[idx_triangs] = 2
    k[idx_sines] = 3
    return k


def engineer_features(b_seq, freq, temp, material):
    """Add engineered features to data set."""
    match b_seq:
        case str():
            raise NotImplementedError("b_seq must be an array-like yet")
        case np.ndarray():
            # check b_seq shapes
            match b_seq.ndim:
                case 1:
                    b_seq = b_seq[np.newaxis, :]
                case 2:
                    pass
                case _:
                    raise ValueError(f"Expected b_seq to have either one or two dimensions, but is has {b_seq.ndim}.")
        case list() | tuple():
            b_seq = np.array(b_seq)
        case _:
            raise ValueError(f"Type of b_seq={type(b_seq)} nut supported. Please provide as np.ndarray or list")

    waveforms = get_waveform_est(b_seq)
    waveforms_df = pd.DataFrame(
        np.zeros((len(waveforms), 4)),
        columns=["wav_other", "wav_square", "wav_triangular", "wav_sine"],
    )
    # one hot encode
    waveform_dummies = pd.get_dummies(waveforms, prefix="wav", dtype=float).rename(
        columns={
            "wav_0": "wav_other",
            "wav_1": "wav_square",
            "wav_2": "wav_triangular",
            "wav_3": "wav_sine",
        }
    )
    for c in waveform_dummies:
        waveforms_df.loc[:, c] = waveform_dummies.loc[:, c]
    ds = pd.DataFrame(b_seq, columns=ALL_B_COLS).assign(
        freq=freq,
        temp=temp,
        material=material,
        **{c: waveforms_df.loc[:, c] for c in waveforms_df},
    )

    dbdt = b_seq[:, 1:] - b_seq[:, :-1]
    b_peak2peak = b_seq.max(axis=1) - b_seq.min(axis=1)

    ds = ds.assign(
        b_peak2peak=b_peak2peak,
        log_peak2peak=np.log(b_peak2peak),
        mean_abs_dbdt=np.mean(np.abs(dbdt), axis=1),
        log_mean_abs_dbdt=np.log(np.mean(np.abs(dbdt), axis=1)),
        sample_time=1 / freq,
    )

    return ds


def construct_tensor_seq2seq(
    df,
    x_cols,
    b_limit,
    h_limit,
    b_limit_pp=None,
    ln_ploss_mean=0,
    ln_ploss_std=1,
    training_data=True,
):
    """
    Generate tensors.

    Shapes as following:
     - For time series tensors (#time steps, #profiles/periods, #features),
     - for scalar tensors (#profiles, #features).
    """
    full_b = df.loc[:, ALL_B_COLS].to_numpy()
    if training_data:
        full_h = df.loc[:, ALL_H_COLS].to_numpy()
    mat = df.iloc[0, :].loc["material"]
    df = df.drop(columns=[c for c in df if c.startswith(("H_t_", "B_t_", "material"))])
    assert len(df) > 0, "empty dataframe error"
    # put freq on first place since Architecture expects it there
    x_cols.insert(0, x_cols.pop(x_cols.index("freq")))
    X = df.loc[:, x_cols].astype(np.float32)

    # normalization
    full_b /= b_limit
    if training_data:
        full_h /= h_limit
    orig_freq = X.loc[:, ["freq"]].copy().to_numpy()
    X.loc[:, ["temp", "freq"]] /= np.array([75.0, FREQ_SCALE], dtype=np.float32)
    X.loc[:, "freq"] = np.log(X.freq)
    other_cols = [c for c in x_cols if c not in ["temp", "freq"] and not c.startswith("wav_")]
    for other_col in other_cols:
        X.loc[:, other_col] /= NORM_DENOM[mat][other_col]

    if training_data:
        # add p loss as target (only used as target when predicting p loss directly), must be last column
        X = X.assign(ln_ploss=(np.log(df.ploss) - ln_ploss_mean) / ln_ploss_std)
    # tensor list
    tens_l = []
    if b_limit_pp is not None:
        # add another B curve with different normalization
        per_profile_scaled_b = full_b * b_limit / b_limit_pp
        # add timeseries derivatives
        b_deriv = np.empty((full_b.shape[0], full_b.shape[1] + 2))
        b_deriv[:, 1:-1] = per_profile_scaled_b
        b_deriv[:, 0] = per_profile_scaled_b[:, -1]
        b_deriv[:, -1] = per_profile_scaled_b[:, 0]
        b_deriv = np.gradient(b_deriv, axis=1) * orig_freq
        b_deriv_sq = np.gradient(b_deriv, axis=1) * orig_freq
        b_deriv = b_deriv[:, 1:-1]
        b_deriv_sq = b_deriv_sq[:, 1:-1]
        tantan_b = -np.tan(0.9 * np.tan(per_profile_scaled_b)) / 6  # tan-tan feature
        tens_l += [
            torch.tensor(per_profile_scaled_b.T[..., np.newaxis], dtype=torch.float32),
            torch.tensor(
                b_deriv.T[..., np.newaxis] / NORM_DENOM[mat]["b_deriv"],
                dtype=torch.float32,
            ),
            torch.tensor(
                b_deriv_sq.T[..., np.newaxis] / NORM_DENOM[mat]["b_deriv_sq"],
                dtype=torch.float32,
            ),
            torch.tensor(tantan_b.T[..., np.newaxis], dtype=torch.float32),
        ]
    tens_l += [torch.tensor(full_b.T[..., np.newaxis], dtype=torch.float32)]  # b field is penultimate column
    if training_data:
        tens_l += [
            torch.tensor(full_h.T[..., np.newaxis], dtype=torch.float32),  # target is last column
        ]

    # return ts tensor with shape: (#time steps, #profiles, #features), and scalar tensor with (#profiles, #features)
    return torch.dstack(tens_l), torch.tensor(X.to_numpy(), dtype=torch.float32)


class PaderbornModel:
    """The Paderborn model.

    HARDCORE: H-field and power loss estimation for arbitrary waveforms with residual,
     dilated convolutional neural networks in ferrite cores
    N Förster, W Kirchgässner, T Piepenbrock, O Schweins, O Wallscheid
    arXiv preprint arXiv:2401.11488

    """

    def __init__(self, model_path, material):
        self.model_path = model_path
        self.material = material
        self.mdl = torch.jit.load(model_path)
        self.mdl.eval()
        assert (
            material in MAT_CONST_H_MAX and material in MAT_CONST_B_MAX
        ), f"Requested material '{material}' is not supported"
        self.b_limit = MAT_CONST_B_MAX[material]
        self.h_limit = MAT_CONST_H_MAX[material]
        self.predicts_p_directly = True

    def __call__(self, b_seq, frequency, temperature):
        """Evaluate trajectory and estimate power loss.

        Args
        ----
        b_seq: (X, Y) array_like
            The magnetic flux density array(s) in T. First dimension X describes the batch, the second Y
             the time length (will always be interpolated to 1024 samples)
        frequency: scalar or 1D array-like
            The frequency operation point(s) in Hz
        temperature: scalar or 1D array-like
            The temperature operation point(s) in °C
        
        Return
        ------
        p, h: (X,) np.array, (X, Y) np.ndarray
            The estimated power loss (p) in W/m³ and the estimated magnetic field strength (h) in A/m.
        """
        ds = engineer_features(b_seq, frequency, temperature, self.material)
        # construct tensors
        x_cols = [c for c in ds if c not in ["ploss", "kfold", "material"] and not c.startswith(("B_t_", "H_t_"))]
        b_limit_per_profile = np.abs(ds.loc[:, ALL_B_COLS].to_numpy()).max(axis=1).reshape(-1, 1)
        h_limit = self.h_limit * b_limit_per_profile / self.b_limit
        b_limit_test_fold = self.b_limit
        b_limit_test_fold_pp = b_limit_per_profile
        h_limit_test_fold = h_limit
        with torch.inference_mode():
            val_tensor_ts, val_tensor_scalar = construct_tensor_seq2seq(
                ds,
                x_cols,
                b_limit_test_fold,
                h_limit_test_fold,
                b_limit_pp=b_limit_test_fold_pp,
                training_data=False,
            )

            if self.predicts_p_directly:
                # prepare torch tensors for normalization scales
                b_limit_test_fold_torch = torch.as_tensor(b_limit_test_fold, dtype=torch.float32)
                h_limit_test_fold_torch = torch.as_tensor(h_limit_test_fold, dtype=torch.float32)
                freq_scale_torch = torch.as_tensor(FREQ_SCALE, dtype=torch.float32)

                val_pred_p, val_pred_h = self.mdl(
                    val_tensor_ts.permute(1, 2, 0),
                    val_tensor_scalar,
                    b_limit_test_fold_torch,
                    h_limit_test_fold_torch,
                    freq_scale_torch,
                )
            else:
                val_pred_h = self.mdl(
                    val_tensor_ts.permute(1, 2, 0),
                    val_tensor_scalar,
                ).permute(2, 0, 1)
                val_pred_p = None
            h_pred = val_pred_h.squeeze().cpu().numpy().T * h_limit_test_fold
            if val_pred_p is None:
                p_pred = frequency * np.trapz(h_pred, b_seq, axis=1)
            else:
                p_pred = np.exp(val_pred_p.squeeze().cpu().numpy())
        return p_pred.astype(np.float32), h_pred.astype(np.float32)
